from __future__ import annotations
from typing import Set, FrozenSet, List, Dict, Union, Tuple, Deque, Optional, TypeVar, Any
from ast import AST
from reg_automata import EFA
from common import State, Character, Eps, Emptyset
from copy import deepcopy
from enum import Enum
import enum


class Bin(Enum):
    Union = enum.auto()
    Concat = enum.auto()


class Iter():
    pass


class AST(object):
    def astprint(self) -> str:
        pass


class BinOp(AST):
    def __init__(self, left: AST, op: Bin, right: AST):
        self.left = left
        self.op = op
        self.right = right
        self.token = op

    def astprint(self):
        if self.op == Bin.Union:
            return "(" + self.left.astprint() + "+" + self.right.astprint() + ")"
        else:
            return self.left.astprint() + "." + self.right.astprint()


class IterOp(AST):
    def __init__(self, node: AST):
        self.node = node
        self.op = Iter
        self.token = Iter

    def astprint(self):
        return self.node.astprint() + "*"


class CharNode(AST):
    def __init__(self, token: Union[Character, Eps, Emptyset]):
        self.token = token
        self.value = token.name

    def astprint(self):
        return self.value

class REG:
    def __init__(self, characters: Set[Character], expression: AST):
        self.characters = characters
        self.expression = expression

    def reg_to_efa(self) -> EFA:
        states: Set[State] = set()
        transition: Dict[Tuple[State, AST], Set[State]] = dict()
        init = State("init")
        final = State("final")
        transition[init, self.expression] = {final}

        new_name = 0
        while self.has_ast(transition):
            changed = False
            for (state, ast) in transition:

                for dest_state in transition[state, ast]:

                    if ast.token == Bin.Union:
                        self.transition_add(transition, state, ast.left, dest_state)
                        self.transition_add(transition, state, ast.right, dest_state)
                        changed = True

                    elif ast.token == Bin.Concat:
                        new_state = State(str(new_name))
                        states.add(new_state)
                        new_name += 1
                        self.transition_add(transition, state, ast.left, new_state)
                        self.transition_add(transition, new_state, ast.right, dest_state)
                        changed = True

                    elif ast.token == Iter:
                        new_state = State(str(new_name))
                        states.add(new_state)
                        new_name += 1
                        self.transition_add(transition, state, CharNode(Eps()),  new_state)
                        self.transition_add(transition, new_state, ast.node,  new_state)
                        self.transition_add(transition, new_state, CharNode(Eps()),  dest_state)
                        changed = True

                if changed:
                    del transition[state, ast]
                    break

        new_transition: Dict[Tuple[State, Character], State] = dict()
        for (state, char) in transition:
            new_transition[state, char.token] = transition[state, char]

        return EFA(states, self.characters, new_transition, init, {final})

    def transition_add(self, transition, state, ast, dest_state):
        if (state, ast) in transition:
            transition[state, ast].add(dest_state)
        else:
            transition[state, ast] = {dest_state}

    def has_ast(self, transition: Dict[Tuple[State, AST], State]):
        for (state, ast) in transition:
            if type(ast) != CharNode:
                return True
        return False